# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import shutil
import tempfile
import unittest
from unittest import mock
import fsspec
from pyglove.core.io import file_system


class StdFileSystemTest(unittest.TestCase):

  def test_file(self):
    tmp_dir = tempfile.mkdtemp()
    fs = file_system.StdFileSystem()

    file1 = os.path.join(tmp_dir, 'file1')
    with self.assertRaises(FileNotFoundError):
      fs.open(file1)

    with fs.open(file1, 'w') as f:
      f.write('hello\npyglove')

    with fs.open(file1, 'r') as f:
      self.assertEqual(f.readline(), 'hello\n')
      self.assertEqual(f.tell(), 6)
      self.assertEqual(f.seek(8, 0), 8)
      self.assertEqual(f.read(), 'glove')

    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.exists('/not-exist-dir'))
    fs.rm(file1)
    self.assertFalse(fs.exists(file1))

  def test_file_system(self):
    tmp_dir = tempfile.mkdtemp()
    fs = file_system.StdFileSystem()

    # Create a directory.
    dir_a = os.path.join(tmp_dir, 'a')
    fs.mkdir(dir_a)
    self.assertTrue(fs.exists(dir_a))
    self.assertTrue(fs.isdir(dir_a))
    self.assertEqual(len(fs.listdir(dir_a)), 0)

    # Create a file under the directory.
    file1 = os.path.join(dir_a, 'file1')
    self.assertFalse(fs.exists(file1))
    with fs.open(file1, 'w') as f:
      f.write('hello')
    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.isdir(file1))

    fs.mkdirs(os.path.join(dir_a, 'b/c/d'))
    self.assertEqual(sorted(fs.listdir(dir_a)), ['b', 'file1'])   # pylint: disable=g-generic-assert

    # Test rm.
    with self.assertRaises(FileNotFoundError):
      fs.rm(os.path.join(dir_a, 'file2'))
    with self.assertRaises((IsADirectoryError, PermissionError)):
      fs.rm(os.path.join(dir_a, 'b'))

    # Test rmdir.
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'c'))
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'file2'))
    fs.rmdir(os.path.join(dir_a, 'b/c/d'))

    # Test rmdirs.
    fs.rmdirs(os.path.join(dir_a, 'b/c'))
    self.assertEqual(sorted(fs.listdir(dir_a)), ['file1'])   # pylint: disable=g-generic-assert

  def test_rename(self):
    tmp_dir = tempfile.mkdtemp()
    fs = file_system.StdFileSystem()

    _ = fs.mkdirs(os.path.join(tmp_dir, 'a/b'))
    file_foo = os.path.join(tmp_dir, 'a/foo.txt')
    file_bar = os.path.join(tmp_dir, 'a/bar.txt')

    with fs.open(file_foo, 'w') as f:
      f.write('foo')
    with fs.open(file_bar, 'w') as f:
      f.write('bar')

    # Rename file to a new name.
    file_foo_new = os.path.join(tmp_dir, 'a/foo-new.txt')
    fs.rename(file_foo, file_foo_new)
    self.assertFalse(fs.exists(file_foo))
    self.assertTrue(fs.exists(file_foo_new))

    # Rename file to an existing file name.
    fs.rename(file_foo_new, file_bar)
    self.assertFalse(fs.exists(file_foo_new))
    with fs.open(file_bar, 'r') as f:
      self.assertEqual(f.read(), 'foo')

    # Rename directory to a new name.
    dir_b = os.path.join(tmp_dir, 'a/b')
    dir_c = os.path.join(tmp_dir, 'a/c')
    fs.rename(dir_b, dir_c)
    self.assertFalse(fs.exists(dir_b))
    self.assertTrue(fs.exists(dir_c))
    self.assertTrue(fs.isdir(dir_c))

    # Rename directory to an existing empty directory.
    dir_d = os.path.join(tmp_dir, 'a/d')
    fs.mkdirs(dir_d)
    fs.rename(dir_c, dir_d)
    self.assertFalse(fs.exists(dir_c))
    self.assertTrue(fs.exists(dir_d))

    # Rename directory to a non-empty directory.
    dir_x = os.path.join(tmp_dir, 'x')
    dir_a = os.path.join(tmp_dir, 'a')
    fs.mkdirs(os.path.join(dir_x, 'y'))
    with self.assertRaises(OSError):
      fs.rename(dir_a, dir_x)
    self.assertTrue(fs.exists(dir_a))
    self.assertTrue(fs.exists(os.path.join(dir_x, 'y')))

    # Errors
    dir_u = os.path.join(tmp_dir, 'u')
    dir_u_v = os.path.join(dir_u, 'v')
    file_u_a = os.path.join(dir_u, 'a.txt')
    fs.mkdirs(dir_u_v)
    with fs.open(file_u_a, 'w') as f:
      f.write('a')

    with self.assertRaises((OSError, NotADirectoryError)):
      fs.rename(dir_u, file_u_a)

    with self.assertRaises(IsADirectoryError):
      fs.rename(file_u_a, dir_u_v)

    with self.assertRaises(FileNotFoundError):
      fs.rename(
          os.path.join(tmp_dir, 'non-existent'),
          os.path.join(tmp_dir, 'y')
      )

  def test_copy(self):
    tmp_dir = tempfile.mkdtemp()
    fs = file_system.StdFileSystem()
    foo_txt = os.path.join(tmp_dir, 'foo.txt')
    bar_txt = os.path.join(tmp_dir, 'bar.txt')
    sub_dir = os.path.join(tmp_dir, 'sub')
    sub_foo_txt = os.path.join(sub_dir, 'foo.txt')

    with fs.open(foo_txt, 'w') as f:
      f.write('hello')
    fs.copy(foo_txt, bar_txt)
    with fs.open(bar_txt) as f:
      self.assertEqual(f.read(), 'hello')
    fs.mkdir(sub_dir)
    fs.copy(foo_txt, sub_dir)
    with fs.open(sub_foo_txt) as f:
      self.assertEqual(f.read(), 'hello')
    with fs.open(foo_txt, 'w') as f:
      f.write('overwrite')
    fs.copy(foo_txt, bar_txt)
    with fs.open(bar_txt) as f:
      self.assertEqual(f.read(), 'overwrite')
    shutil.rmtree(tmp_dir)


class MemoryFileSystemTest(unittest.TestCase):

  def test_file(self):
    fs = file_system.MemoryFileSystem()

    file1 = os.path.join('/mem', 'file1')
    with self.assertRaises(FileNotFoundError):
      fs.open(file1)

    with self.assertRaisesRegex(ValueError, 'Unsupported path'):
      fs.open(1)

    with fs.open(file1, 'w') as f:
      f.write('hello\npyglove')

    with fs.open(file1, 'r') as f:
      self.assertEqual(f.readline(), 'hello\n')
      self.assertEqual(f.tell(), 6)
      self.assertEqual(f.seek(8, 0), 8)
      self.assertEqual(f.read(), 'glove')

    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.exists('/not-exist-dir'))
    fs.rm(file1)
    self.assertFalse(fs.exists(file1))

  def test_file_system(self):
    fs = file_system.MemoryFileSystem()

    # Create a directory.
    dir_a = os.path.join('/mem', 'a')
    fs.mkdir(dir_a)
    self.assertTrue(fs.exists(dir_a))
    self.assertTrue(fs.isdir(dir_a))
    self.assertEqual(len(fs.listdir(dir_a)), 0)

    with self.assertRaises(FileExistsError):
      fs.mkdir(dir_a)

    with self.assertRaises(FileNotFoundError):
      fs.mkdir('/mem/b/c')

    with self.assertRaises(IsADirectoryError):
      fs.open(dir_a)

    # Create a file under the directory.
    file1 = os.path.join(dir_a, 'file1')
    self.assertFalse(fs.exists(file1))
    with fs.open(file1, 'w') as f:
      f.write('hello')
      f.flush()
    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.isdir(file1))

    # Make dirs.
    fs.mkdirs(os.path.join(dir_a, 'b/c/d'))

    fs.mkdirs(os.path.join(dir_a, 'b/c/d'))
    with self.assertRaises(FileExistsError):
      fs.mkdirs(os.path.join(dir_a, 'b/c/d'), exist_ok=False)
    with self.assertRaises(NotADirectoryError):
      fs.mkdirs(os.path.join(file1, 'e'))

    with self.assertRaises(FileNotFoundError):
      fs.listdir('/mem/not-exist')
    self.assertEqual(sorted(fs.listdir(dir_a)), ['b', 'file1'])   # pylint: disable=g-generic-assert

    # Test rm.
    with self.assertRaises(FileNotFoundError):
      fs.rm(os.path.join(dir_a, 'file2'))
    with self.assertRaises(IsADirectoryError):
      fs.rm(os.path.join(dir_a, 'b'))

    # Test rmdir.
    with self.assertRaisesRegex(OSError, 'Directory not empty'):
      fs.rmdir(dir_a)
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'c'))
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'file2'))
    with self.assertRaises(NotADirectoryError):
      fs.rmdir(os.path.join(dir_a, 'file1'))
    fs.rmdir(os.path.join(dir_a, 'b/c/d'))

    # Test rmdirs.
    with self.assertRaisesRegex(OSError, 'Directory not empty'):
      fs.rmdirs(dir_a)

    with self.assertRaises(NotADirectoryError):
      fs.rmdirs(os.path.join(file1, 'a'))

    with self.assertRaises(FileNotFoundError):
      fs.rmdirs('a/b/d')

    fs.rmdirs(os.path.join(dir_a, 'b/c'))
    self.assertEqual(fs.listdir(dir_a), ['file1'])   # pylint: disable=g-generic-assert

  def test_glob(self):
    fs = file_system.MemoryFileSystem()
    fs.mkdirs('/mem/a/b/c')
    with fs.open('/mem/a/foo.txt', 'w') as f:
      f.write('foo')
    with fs.open('/mem/a/bar.json', 'w') as f:
      f.write('bar')
    with fs.open('/mem/a/b/baz.txt', 'w') as f:
      f.write('baz')

    self.assertEqual(
        sorted(fs.glob('/mem/a/*')),
        ['/mem/a/b', '/mem/a/b/baz.txt', '/mem/a/b/c',
         '/mem/a/bar.json', '/mem/a/foo.txt'])
    self.assertEqual(
        sorted(fs.glob('/mem/a/*.txt')),
        ['/mem/a/b/baz.txt', '/mem/a/foo.txt'])
    self.assertEqual(
        sorted(fs.glob('/mem/a/b/*')),
        ['/mem/a/b/baz.txt', '/mem/a/b/c'])
    self.assertEqual(fs.glob('/mem/a/b/*.txt'), ['/mem/a/b/baz.txt'])
    self.assertEqual(fs.glob('/mem/a/b/c/*'), [])
    self.assertEqual(fs.glob('/mem/a/???.txt'), ['/mem/a/foo.txt'])
    self.assertEqual(fs.glob('/mem/a/bar.*'), ['/mem/a/bar.json'])
    self.assertEqual(
        sorted(fs.glob('/mem/a/*.*')),
        ['/mem/a/b/baz.txt', '/mem/a/bar.json', '/mem/a/foo.txt'])

  def test_rename(self):
    fs = file_system.MemoryFileSystem()
    fs.mkdirs('/mem/a/b')
    with fs.open('/mem/a/foo.txt', 'w') as f:
      f.write('foo')
    with fs.open('/mem/a/bar.txt', 'w') as f:
      f.write('bar')

    # Rename file to a new name.
    fs.rename('/mem/a/foo.txt', '/mem/a/foo-new.txt')
    self.assertFalse(fs.exists('/mem/a/foo.txt'))
    self.assertTrue(fs.exists('/mem/a/foo-new.txt'))

    # Rename file to an existing file name.
    fs.rename('/mem/a/foo-new.txt', '/mem/a/bar.txt')
    self.assertFalse(fs.exists('/mem/a/foo-new.txt'))
    with fs.open('/mem/a/bar.txt', 'r') as f:
      self.assertEqual(f.read(), 'foo')

    # Rename directory to a new name.
    fs.rename('/mem/a/b', '/mem/a/c')
    self.assertFalse(fs.exists('/mem/a/b'))
    self.assertTrue(fs.exists('/mem/a/c'))
    self.assertTrue(fs.isdir('/mem/a/c'))

    # Rename directory to an existing empty directory.
    fs.mkdirs('/mem/a/d')
    fs.rename('/mem/a/c', '/mem/a/d')
    self.assertFalse(fs.exists('/mem/a/c'))
    self.assertTrue(fs.exists('/mem/a/d'))

    # Rename directory to a non-empty directory.
    fs.mkdirs('/mem/x/y')
    with self.assertRaisesRegex(OSError, "Directory not empty: '/mem/x'"):
      fs.rename('/mem/a', '/mem/x')
    self.assertTrue(fs.exists('/mem/a'))
    self.assertTrue(fs.exists('/mem/x/y'))

    # Errors
    fs.mkdirs('/mem/u/v')
    with fs.open('/mem/u/a.txt', 'w') as f:
      f.write('a')

    with self.assertRaisesRegex(
        OSError, "Cannot move directory '/mem/u' to a subdirectory of itself"):
      fs.rename('/mem/u', '/mem/u/v/w')

    with self.assertRaisesRegex(
        NotADirectoryError,
        "Cannot rename directory '/mem/u' to non-directory '/mem/u/a.txt'"):
      fs.rename('/mem/u', '/mem/u/a.txt')

    with self.assertRaisesRegex(
        IsADirectoryError,
        "Cannot rename non-directory '/mem/u/a.txt' to directory '/mem/u/v'"):
      fs.rename('/mem/u/a.txt', '/mem/u/v')

    with self.assertRaises(FileNotFoundError):
      fs.rename('/mem/non-existent', '/mem/y')

  def test_copy(self):
    fs = file_system.MemoryFileSystem()
    fs.mkdirs('/mem/a')
    with fs.open('/mem/a/foo.txt', 'w') as f:
      f.write('hello')
    fs.copy('/mem/a/foo.txt', '/mem/a/bar.txt')
    self.assertEqual(fs.open('/mem/a/bar.txt').read(), 'hello')
    fs.mkdir('/mem/b')
    fs.copy('/mem/a/foo.txt', '/mem/b')
    self.assertEqual(fs.open('/mem/b/foo.txt').read(), 'hello')
    with fs.open('/mem/a/foo.txt', 'w') as f:
      f.write('overwrite')
    fs.copy('/mem/a/foo.txt', '/mem/a/bar.txt')
    self.assertEqual(fs.open('/mem/a/bar.txt').read(), 'overwrite')

    # Test exceptions
    with self.assertRaises(FileNotFoundError):
      fs.copy('/mem/non-existent', '/mem/y')
    with self.assertRaisesRegex(IsADirectoryError, '/mem/a'):
      fs.copy('/mem/a', '/mem/y')

    fs.mkdirs('/mem/c/foo.txt')
    with self.assertRaisesRegex(IsADirectoryError, '/mem/c/foo.txt'):
      fs.copy('/mem/a/foo.txt', '/mem/c')


class FileIoApiTest(unittest.TestCase):

  def test_standard_filesystem(self):
    file1 = os.path.join(tempfile.mkdtemp(), 'file1')
    with self.assertRaises(FileNotFoundError):
      file_system.readfile(file1)
    self.assertIsNone(file_system.readfile(file1, nonexist_ok=True))
    with file_system.open(file1, 'w') as f:
      self.assertIsInstance(f, file_system.StdFile)
      f.write('foo')
    self.assertTrue(file_system.path_exists(file1))
    self.assertEqual(file_system.readfile(file1), 'foo')
    file_system.writefile(file1, 'bar')
    self.assertEqual(file_system.readfile(file1), 'bar')
    self.assertFalse(file_system.isdir(file1))
    file_system.rm(file1)
    self.assertFalse(file_system.path_exists(file1))

    dir1 = os.path.join(tempfile.mkdtemp(), 'dir1')
    file_system.mkdir(dir1)
    self.assertEqual(file_system.listdir(dir1), [])
    file_system.mkdirs(os.path.join(dir1, 'a/b/c'))
    file2 = os.path.join(dir1, 'file2')
    file_system.writefile(file2, 'baz')
    self.assertEqual(sorted(file_system.listdir(dir1)), ['a', 'file2'])  # pylint: disable=g-generic-assert
    self.assertEqual(    # pylint: disable=g-generic-assert
        sorted(file_system.listdir(dir1, fullpath=True)),
        [os.path.join(dir1, 'a'), os.path.join(dir1, 'file2')]
    )
    file_system.rmdir(os.path.join(dir1, 'a/b/c'))
    file_system.rmdirs(os.path.join(dir1, 'a/b'))
    self.assertTrue(file_system.path_exists(dir1))
    self.assertEqual(sorted(file_system.listdir(dir1)), ['file2'])  # pylint: disable=g-generic-assert
    file_system.rm(file2)
    self.assertFalse(file_system.path_exists(file2))

    # Test glob with standard file system.
    glob_dir = os.path.join(tempfile.mkdtemp(), 'glob')
    file_system.mkdirs(os.path.join(glob_dir, 'a/b'))
    file_system.writefile(os.path.join(glob_dir, 'a/foo.txt'), 'foo')
    self.assertEqual(
        sorted(file_system.glob(os.path.join(glob_dir, 'a/*'))),
        [os.path.join(glob_dir, 'a/b'), os.path.join(glob_dir, 'a/foo.txt')])

  def test_memory_filesystem(self):
    file1 = pathlib.Path('/mem/file1')
    with self.assertRaises(FileNotFoundError):
      file_system.readfile(file1)
    self.assertIsNone(file_system.readfile(file1, nonexist_ok=True))
    with file_system.open(file1, 'w') as f:
      self.assertIsInstance(f, file_system.MemoryFile)
      f.write('foo')
      f.flush()
    self.assertTrue(file_system.path_exists(file1))
    self.assertEqual(file_system.readfile(file1), 'foo')
    file_system.writefile(file1, 'bar')
    self.assertEqual(file_system.readfile(file1), 'bar')
    self.assertFalse(file_system.isdir(file1))
    file_system.rm(file1)
    self.assertFalse(file_system.path_exists(file1))

    dir1 = os.path.join('/mem/dir1')
    file_system.mkdir(dir1)
    self.assertEqual(file_system.listdir(dir1), [])
    file_system.mkdirs(os.path.join(dir1, 'a/b/c'))
    file2 = os.path.join(dir1, 'file2')
    file_system.writefile(file2, 'baz')
    self.assertEqual(sorted(file_system.listdir(dir1)), ['a', 'file2'])  # pylint: disable=g-generic-assert
    file_system.rmdir(os.path.join(dir1, 'a/b/c'))
    file_system.rmdirs(os.path.join(dir1, 'a/b'))
    self.assertTrue(file_system.path_exists(dir1))
    self.assertEqual(sorted(file_system.listdir(dir1)), ['file2'])  # pylint: disable=g-generic-assert
    file_system.rm(file2)
    self.assertFalse(file_system.path_exists(file2))

    # Test glob with memory file system.
    file_system.mkdirs('/mem/g/a/b')
    file_system.writefile('/mem/g/a/foo.txt', 'foo')
    file_system.rename('/mem/g/a/foo.txt', '/mem/g/a/foo2.txt')
    file_system.writefile('/mem/g/a/b/bar.txt', 'bar')
    self.assertEqual(
        sorted(file_system.glob('/mem/g/a/*')),
        ['/mem/g/a/b', '/mem/g/a/b/bar.txt', '/mem/g/a/foo2.txt'])


class FsspecFileSystemTest(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.fs = fsspec.filesystem('memory')
    self.fs.pipe('memory:///a/b/c', b'abc')
    self.fs.pipe('memory:///a/b/d', b'abd')
    self.fs.mkdir('memory:///a/e')
    self.tmp_dir = tempfile.mkdtemp()

  def tearDown(self):
    super().tearDown()
    fsspec.filesystem('memory').rm('/', recursive=True)
    shutil.rmtree(self.tmp_dir)

  def test_read_file(self):
    self.assertEqual(file_system.readfile('memory:///a/b/c', mode='rb'), b'abc')
    with file_system.open('memory:///a/b/d', 'rb') as f:
      self.assertEqual(f.read(), b'abd')

  def test_fsspec_file_ops(self):
    file_system.writefile('memory:///f', b'hello\nworld\n', mode='wb')
    with file_system.open('memory:///f', 'rb') as f:
      self.assertIsInstance(f, file_system.FsspecFile)
      self.assertEqual(f.readline(), b'hello\n')
      self.assertEqual(f.tell(), 6)
      self.assertEqual(f.seek(8), 8)
      self.assertEqual(f.read(), b'rld\n')
      f.flush()

  def test_write_file(self):
    file_system.writefile('memory:///a/b/e', b'abe', mode='wb')
    self.assertTrue(self.fs.exists('memory:///a/b/e'))
    self.assertEqual(self.fs.cat('memory:///a/b/e'), b'abe')

  def test_exists(self):
    self.assertTrue(file_system.path_exists('memory:///a/b/c'))
    self.assertFalse(file_system.path_exists('memory:///a/b/nonexist'))

  def test_isdir(self):
    self.assertTrue(file_system.isdir('memory:///a/b'))
    self.assertTrue(file_system.isdir('memory:///a/e'))
    self.assertFalse(file_system.isdir('memory:///a/b/c'))

  def test_listdir(self):
    self.assertCountEqual(file_system.listdir('memory:///a'), ['b', 'e'])
    self.assertCountEqual(file_system.listdir('memory:///a/b'), ['c', 'd'])

  def test_glob(self):
    self.assertCountEqual(
        file_system.glob('memory:///a/b/*'),
        ['memory:///a/b/c', 'memory:///a/b/d']
    )

  def test_mkdir(self):
    file_system.mkdir('memory:///a/f')
    self.assertTrue(self.fs.isdir('memory:///a/f'))

  def test_mkdirs(self):
    file_system.mkdirs('memory:///g/h/i')
    self.assertTrue(self.fs.isdir('memory:///g/h/i'))

  def test_rm(self):
    file_system.rm('memory:///a/b/c')
    self.assertFalse(self.fs.exists('memory:///a/b/c'))

  def test_rename(self):
    file_system.rename('memory:///a/b/c', 'memory:///a/b/c_new')
    self.assertFalse(self.fs.exists('memory:///a/b/c'))
    self.assertTrue(self.fs.exists('memory:///a/b/c_new'))
    with self.assertRaisesRegex(ValueError, 'Rename across different'):
      file_system.rename('memory:///a/b/c_new', 'file:///a/b/c_d')

  def test_chmod(self):
    mock_fs = mock.Mock()
    mock_fs.chmod = mock.Mock()
    with mock.patch('fsspec.core.url_to_fs', return_value=(mock_fs, 'path')):
      file_system.chmod('protocol:///path', 0o777)
      mock_fs.chmod.assert_called_once_with('path', 0o777)

  def test_rmdir(self):
    file_system.rmdir('memory:///a/e')
    self.assertFalse(self.fs.exists('memory:///a/e'))

  def test_rmdirs(self):
    file_system.mkdirs('memory:///x/y/z')
    self.assertTrue(file_system.isdir('memory:///x/y/z'))
    file_system.rmdirs('memory:///x')
    self.assertFalse(file_system.path_exists('memory:///x'))

  def test_copy(self):
    # same FS copy
    file_system.copy('memory:///a/b/c', 'memory:///a/f')
    self.assertEqual(file_system.readfile('memory:///a/f', mode='rb'), b'abc')

    # same FS copy to dir
    file_system.copy('memory:///a/b/d', 'memory:///a/e')
    self.assertEqual(file_system.readfile('memory:///a/e/d', mode='rb'), b'abd')

    # cross FS copy: memory to local
    local_path = os.path.join(self.tmp_dir, 'test.txt')
    file_system.copy('memory:///a/b/c', f'file://{local_path}')
    self.assertEqual(file_system.readfile(local_path, mode='rb'), b'abc')

    # cross FS copy: local to memory
    file_system.copy(f'file://{local_path}', 'memory:///a/g')
    self.assertEqual(file_system.readfile('memory:///a/g', mode='rb'), b'abc')

  def test_fsspec_uri_catcher(self):
    with mock.patch.object(
        file_system.FsspecFileSystem, 'exists', return_value=True
    ) as mock_fsspec_exists:
      # We use a protocol that is not registered in
      # fsspec.available_protocols() to make sure _FsspecUriCatcher is used.
      self.assertTrue(file_system.path_exists('some-proto://foo'))
      mock_fsspec_exists.assert_called_once_with('some-proto://foo')

    # For full coverage of _FsspecUriCatcher.get_fs returning StdFileSystem,
    # we need to test a non-URI path that doesn't match other prefixes.
    # We mock StdFileSystem.exists to check if it's called.
    with mock.patch.object(
        file_system.StdFileSystem, 'exists', return_value=True
    ) as mock_std_exists:
      self.assertTrue(file_system.path_exists('/foo/bar/baz'))
      mock_std_exists.assert_called_once_with('/foo/bar/baz')

    with mock.patch.object(
        file_system.FsspecFileSystem, 'rename', return_value=None
    ) as mock_fsspec_rename:
      file_system.rename('some-proto://foo', 'some-proto://bar')
      mock_fsspec_rename.assert_called_once_with(
          'some-proto://foo', 'some-proto://bar'
      )

    with mock.patch.object(
        file_system.FsspecFileSystem, 'copy', return_value=None
    ) as mock_fsspec_copy:
      file_system.copy('some-proto://foo', 'some-proto://bar')
      mock_fsspec_copy.assert_called_once_with(
          'some-proto://foo', 'some-proto://bar'
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Injecting and manipulating values through context managers."""

import contextlib
import dataclasses
import inspect
import threading
from typing import Any, Callable, ContextManager, Iterator, Optional

from pyglove.core.utils import missing

RAISE_IF_HAS_ERROR = (missing.MISSING_VALUE,)
_TLS_KEY_CONTEXTUAL_OVERRIDES = '__contextual_overrides__'
_global_contextual_overrides = threading.local()


@dataclasses.dataclass(frozen=True)
class ContextualOverride:
  """Value marker for contextual override for an attribute."""

  # Overridden value.
  value: Any

  # If True, this override will apply to both current scope and nested scope,
  # meaning current `pg.contextual_override` will take precedence over all
  # nested `pg.contextual_override` on this attribute.
  cascade: bool = False

  # If True, this override will apply to attributes that already have values.
  override_attrs: bool = False


def contextual_override(
    *,
    cascade: bool = False,
    override_attrs: bool = False,
    **variables,
) -> ContextManager[dict[str, ContextualOverride]]:
  """Context manager to provide contextual values under a scope.

  Please be aware that contextual value override are per-thread. If you want
  to propagate the contextual value override to other threads, please obtain
  a wrapper function for a user function using
  `pg.with_contextual_override(func)`.

  Args:
    cascade: If True, this override will apply to both current scope and nested
      scope, meaning that this `pg.contextual_override` will take precedence
      over all nested `pg.contextual_override` on the overriden variables.
    override_attrs: If True, this override will apply to attributes that already
      have values. Otherwise overridden variables will only be used for
      contextual attributes whose values are not present.
    **variables: Key/values as override for contextual attributes.

  Returns:
    A dict of attribute names to their contextual overrides.
  """
  vs = {}
  for k, v in variables.items():
    if not isinstance(v, ContextualOverride):
      v = ContextualOverride(v, cascade, override_attrs)
    vs[k] = v
  return contextual_scope(_global_contextual_overrides, **vs)


def with_contextual_override(func: Callable[..., Any]) -> Callable[..., Any]:
  """Wraps a user function with the access to the current contextual override.

  The wrapped function can be called from another thread.

  Args:
    func: The user function to be wrapped.

  Returns:
    A wrapper function that have the access to the current contextual override,
    which can be called from another thread.
  """
  with contextual_override() as current_context:
    pass

  if inspect.iscoroutinefunction(func):
    async def _func(*args, **kwargs) -> Any:
      with contextual_override(**current_context):
        return await func(*args, **kwargs)
  else:
    def _func(*args, **kwargs) -> Any:
      with contextual_override(**current_context):
        return func(*args, **kwargs)
  return _func


def get_contextual_override(var_name: str) -> Optional[ContextualOverride]:
  """Returns the overriden contextual value in current scope."""
  return get_scoped_value(_global_contextual_overrides, var_name)


def contextual_value(var_name: str, default: Any = RAISE_IF_HAS_ERROR) -> Any:
  """Returns the value of a variable defined in `pg.contextual_override`."""
  override = get_contextual_override(var_name)
  if override is None:
    if default == RAISE_IF_HAS_ERROR:
      raise KeyError(f'{var_name!r} does not exist in current context.')
    return default
  return override.value


def all_contextual_values() -> dict[str, Any]:
  """Returns all values provided from `pg.contextual_override` in scope."""
  overrides = getattr(
      _global_contextual_overrides, _TLS_KEY_CONTEXTUAL_OVERRIDES, {}
  )
  return {k: v.value for k, v in overrides.items()}


@contextlib.contextmanager
def contextual_scope(
    tls: threading.local, **variables
) -> Iterator[dict[str, ContextualOverride]]:
  """Context manager to set variables within a scope."""
  previous_values = getattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, {})
  current_values = dict(previous_values)
  for k, v in variables.items():
    old_v = current_values.get(k, None)
    if old_v and old_v.cascade:
      v = old_v
    current_values[k] = v
  try:
    setattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, current_values)
    yield current_values
  finally:
    setattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, previous_values)


def get_scoped_value(
    tls: threading.local, var_name: str, default: Any = None
) -> ContextualOverride:
  """Gets the value for requested variable from current scope."""
  scoped_values = getattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, {})
  return scoped_values.get(var_name, default)




# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pluggable file system."""

import abc
import fnmatch
import glob as std_glob
import io
import os
import re
import shutil
from typing import Any, Literal, Optional, Union


class File(metaclass=abc.ABCMeta):
  """Interface for a file."""

  @abc.abstractmethod
  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    """Reads the content of the file."""

  @abc.abstractmethod
  def readline(self) -> Union[str, bytes]:
    """Reads the next line."""

  @abc.abstractmethod
  def write(self, content: Union[str, bytes]) -> None:
    """Writes the content of the file."""

  @abc.abstractmethod
  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    """Changes the current position of the file.

    Args:
      offset: Offset from the position to a reference point.
      whence: The reference point, with 0 meaning the beginning of the file,
        1 meaning the current position, or 2 meaning the end of the file.

    Returns:
      The position from the beginning of the file.
    """

  @abc.abstractmethod
  def tell(self) -> int:
    """Returns the current position of the file."""

  @abc.abstractmethod
  def flush(self) -> None:
    """Flushes the written content to the storage."""

  @abc.abstractmethod
  def close(self) -> None:
    """Closes the file."""

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    del exc_type, exc_value, traceback
    self.close()


class FileSystem(metaclass=abc.ABCMeta):
  """Interface for a file system."""

  @abc.abstractmethod
  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    """Opens a file with a path."""

  @abc.abstractmethod
  def chmod(self, path: Union[str, os.PathLike[str]], mode: int) -> None:
    """Changes the permission of a file."""

  @abc.abstractmethod
  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    """Returns True if a path exists."""

  @abc.abstractmethod
  def glob(self, pattern: Union[str, os.PathLike[str]]) -> list[str]:
    """Lists all files or sub-directories."""

  @abc.abstractmethod
  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    """Lists all files or sub-directories."""

  @abc.abstractmethod
  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    """Returns True if a path is a directory."""

  @abc.abstractmethod
  def mkdir(
      self, path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    """Makes a directory based on a path."""

  @abc.abstractmethod
  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = True,
  ) -> None:
    """Makes a directory chain based on a path."""

  @abc.abstractmethod
  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    """Removes a file based on a path."""

  @abc.abstractmethod
  def rename(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    """Renames a file based on a path."""

  @abc.abstractmethod
  def rmdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    """Removes a directory based on a path."""

  @abc.abstractmethod
  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    """Removes a directory chain based on a path."""

  @abc.abstractmethod
  def copy(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    """Copies a file to a new path."""


def resolve_path(path: Union[str, os.PathLike[str]]) -> str:
  if isinstance(path, str):
    return path
  elif hasattr(path, '__fspath__'):
    return path.__fspath__()
  else:
    raise ValueError(f'Unsupported path: {path!r}.')


#
# The standard file system.
#


class StdFile(File):
  """The standard file."""

  def __init__(self, file_object, path: Union[str, os.PathLike[str]]) -> None:
    super().__init__()
    self._file_object = file_object
    self._path = path

  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    return self._file_object.read(size)

  def readline(self) -> Union[str, bytes]:
    return self._file_object.readline()

  def write(self, content: Union[str, bytes]) -> None:
    self._file_object.write(content)

  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    return self._file_object.seek(offset, whence)

  def tell(self) -> int:
    return self._file_object.tell()

  def flush(self) -> None:
    self._file_object.flush()

  def close(self) -> None:
    self._file_object.close()


class StdFileSystem(FileSystem):
  """The standard file system."""

  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    return StdFile(io.open(path, mode, **kwargs), path)

  def chmod(self, path: Union[str, os.PathLike[str]], mode: int) -> None:
    os.chmod(path, mode)

  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    return os.path.exists(path)

  def glob(self, pattern: Union[str, os.PathLike[str]]) -> list[str]:
    return std_glob.glob(pattern)

  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    return os.listdir(path)

  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    return os.path.isdir(path)

  def mkdir(
      self, path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    os.mkdir(path, mode)

  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = True,
  ) -> None:
    os.makedirs(path, mode, exist_ok)

  def rename(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    os.rename(oldpath, newpath)

  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    os.remove(path)

  def rmdir(self, path: Union[str, os.PathLike[str]]) -> None:  # pytype: disable=signature-mismatch
    os.rmdir(path)

  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    os.removedirs(path)

  def copy(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    shutil.copy(oldpath, newpath)


#
# Memory file system.
#


class MemoryFile(File):
  """Memory file."""

  def __init__(self, buffer: io.IOBase):
    super().__init__()
    self._buffer = buffer
    self._pos = 0

  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    return self._buffer.read(size)

  def readline(self) -> Union[str, bytes]:
    return self._buffer.readline()

  def write(self, content: Union[str, bytes]) -> None:
    self._buffer.write(content)

  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    return self._buffer.seek(offset, whence)

  def tell(self) -> int:
    return self._buffer.tell()

  def flush(self) -> None:
    pass

  def close(self) -> None:
    self.seek(0)


class MemoryFileSystem(FileSystem):
  """The in-memory file system."""

  def __init__(self, prefix: str = '/mem/'):
    super().__init__()
    self._root = {}
    self._prefix = prefix

  def _internal_path(self, path: Union[str, os.PathLike[str]]) -> str:
    return '/' + resolve_path(path).lstrip(self._prefix)

  def _locate(self, path: Union[str, os.PathLike[str]]) -> Any:
    current = self._root
    for x in self._internal_path(path).split('/'):
      if not x:
        continue
      if x not in current:
        return None
      current = current[x]
    return current

  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    file = self._locate(path)
    if isinstance(file, dict):
      raise IsADirectoryError(path)
    if 'w' in mode and file is None:
      parent_dir, name = self._parent_and_name(path)
      if isinstance(parent_dir, dict):
        buffer = io.BytesIO() if 'b' in mode else io.StringIO()
        file = MemoryFile(buffer)
        parent_dir[name] = file

    if file is None:
      raise FileNotFoundError(path)
    return file

  def chmod(self, path: Union[str, os.PathLike[str]], mode: int) -> None:
    # No-op.
    del path, mode

  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    return self._locate(path) is not None

  def glob(self, pattern: Union[str, os.PathLike[str]]) -> list[str]:
    pattern = resolve_path(pattern)
    regex = re.compile(fnmatch.translate(pattern))

    results = []
    def _traverse(node, prefix_parts):
      for k, v in node.items():
        path = self._prefix + '/'.join(prefix_parts + [k])
        if regex.match(path):
          results.append(path)
        if isinstance(v, dict):
          _traverse(v, prefix_parts + [k])
    _traverse(self._root, [])
    return results

  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    d = self._locate(path)
    if not isinstance(d, dict):
      raise FileNotFoundError(path)
    return list(d.keys())

  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    return isinstance(self._locate(path), dict)

  def _parent_and_name(
      self, path: Union[str, os.PathLike[str]]
  ) -> tuple[dict[str, Any], str]:
    path = resolve_path(path)
    rpos = path.rfind('/')
    assert rpos >= 0, path
    name = path[rpos + 1:]
    parent_dir = self._locate(path[:rpos])
    if parent_dir is None:
      raise FileNotFoundError(path)
    return parent_dir, name

  def mkdir(
      self, path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    del mode
    parent_dir, name = self._parent_and_name(path)
    if name in parent_dir:
      raise FileExistsError(path)
    parent_dir[name] = {}

  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = True,
  ) -> None:
    del mode
    current = self._root
    dirs = self._internal_path(path).split('/')
    for i, x in enumerate(dirs):
      if not x:
        continue
      entry = current.get(x)
      if entry is None:
        entry = {}
        current[x] = entry
      elif isinstance(entry, dict) and i == len(dirs) - 1 and not exist_ok:
        raise FileExistsError(path)
      elif not isinstance(entry, dict):
        raise NotADirectoryError(path)
      current = entry

  def rename(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    oldpath_str = resolve_path(oldpath)
    newpath_str = resolve_path(newpath)

    old_parent_dir, old_name = self._parent_and_name(oldpath_str)
    entry = old_parent_dir.get(old_name)

    if entry is None:
      raise FileNotFoundError(oldpath_str)

    new_entry = self._locate(newpath_str)
    if new_entry is not None:
      if isinstance(entry, dict):  # oldpath is dir
        if not isinstance(new_entry, dict):
          raise NotADirectoryError(
              f'Cannot rename directory {oldpath_str!r} '
              f'to non-directory {newpath_str!r}')
        elif new_entry:
          raise OSError(f'Directory not empty: {newpath_str!r}')
        else:
          self.rmdir(newpath_str)
      else:  # oldpath is file
        if isinstance(new_entry, dict):
          raise IsADirectoryError(
              f'Cannot rename non-directory {oldpath_str!r} '
              f'to directory {newpath_str!r}')
        else:
          self.rm(newpath_str)
    elif isinstance(entry, dict) and newpath_str.startswith(oldpath_str + '/'):
      raise OSError(f'Cannot move directory {oldpath_str!r} '
                    f'to a subdirectory of itself {newpath_str!r}')

    new_parent_dir, new_name = self._parent_and_name(newpath_str)
    new_parent_dir[new_name] = entry
    del old_parent_dir[old_name]

  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    parent_dir, name = self._parent_and_name(path)
    entry = parent_dir.get(name)
    if entry is None:
      raise FileNotFoundError(path)
    elif isinstance(entry, dict):
      raise IsADirectoryError(path)
    del parent_dir[name]

  def rmdir(self, path: Union[str, os.PathLike[str]]) -> None:  # pytype: disable=signature-mismatch
    parent_dir, name = self._parent_and_name(path)
    entry = parent_dir.get(name)
    if entry is None:
      raise FileNotFoundError(path)
    elif not isinstance(entry, dict):
      raise NotADirectoryError(path)
    elif entry:
      raise OSError(f'Directory not empty: {path!r}')
    del parent_dir[name]

  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    def _rmdir(dir_dict, subpath: str) -> bool:
      if not subpath:
        if dir_dict:
          raise OSError(f'Directory not empty: {path!r}')
        return True
      subpath = subpath.lstrip('/')
      pos = subpath.find('/')
      if pos < 0:
        pos = len(subpath)
      name = subpath[:pos]
      subpath = subpath[pos + 1:] if pos < len(subpath) else ''
      if name not in dir_dict:
        raise FileNotFoundError(path)
      elif not isinstance(dir_dict[name], dict):
        raise NotADirectoryError(path)
      if _rmdir(dir_dict[name], subpath):
        del dir_dict[name]
      return not dir_dict
    _rmdir(self._root, self._internal_path(path))

  def copy(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    old_file = self._locate(oldpath)
    if old_file is None:
      raise FileNotFoundError(oldpath)
    if not isinstance(old_file, MemoryFile):
      raise IsADirectoryError(oldpath)

    if self.isdir(newpath):
      _, old_file_name = self._parent_and_name(oldpath)
      newpath = resolve_path(newpath)
      newpath = newpath.rstrip('/') + '/' + old_file_name

    # If newpath exists as file, remove it for overwrite.
    if self.exists(newpath) and not self.isdir(newpath):
      self.rm(newpath)
    elif self.isdir(newpath):
      raise IsADirectoryError(newpath)

    old_pos = old_file.tell()
    old_file.seek(0)
    content = old_file.read()
    old_file.seek(old_pos)

    is_binary = isinstance(old_file._buffer, io.BytesIO)  # pylint: disable=protected-access
    with self.open(newpath, 'wb' if is_binary else 'w') as f_new:
      f_new.write(content)


class _FileSystemRegistry:
  """File system registry."""

  def __init__(self):
    self._filesystems = []

  def add(self, prefix, fs: FileSystem) -> None:
    """Gets the file system applicable for path."""
    self._filesystems.append((prefix, fs))
    self._filesystems.sort(key=lambda x: x[0], reverse=True)

  def get(self, path: Union[str, os.PathLike[str]]) -> FileSystem:
    """Gets the file system for a path."""
    path = resolve_path(path)
    for prefix, fs in self._filesystems:
      if path.startswith(prefix):
        return fs
    return StdFileSystem()


_fs = _FileSystemRegistry()


def add_file_system(prefix: str, fs: FileSystem) -> None:
  """Adds a file system with a prefix."""
  _fs.add(prefix, fs)


# Register a memory file system.
add_file_system('/mem/', MemoryFileSystem('/mem/'))


try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  import fsspec
  # pytype: enable=import-error
  # pylint: enable=g-import-not-at-top
except ImportError:
  fsspec = None


class FsspecFile(File):
  """File object based on fsspec."""

  def __init__(self, f):
    self._f = f

  def read(self, size: Optional[int] = None) -> Union[str, bytes]:
    return self._f.read(size)

  def readline(self) -> Union[str, bytes]:
    return self._f.readline()

  def write(self, content: Union[str, bytes]) -> None:
    self._f.write(content)

  def seek(self, offset: int, whence: Literal[0, 1, 2] = 0) -> int:
    return self._f.seek(offset, whence)

  def tell(self) -> int:
    return self._f.tell()

  def flush(self) -> None:
    self._f.flush()

  def close(self) -> None:
    self._f.close()


class FsspecFileSystem(FileSystem):
  """File system based on fsspec."""

  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    assert fsspec is not None, '`fsspec` is not installed.'
    return FsspecFile(fsspec.open(path, mode, **kwargs).open())

  def chmod(self, path: Union[str, os.PathLike[str]], mode: int) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    if hasattr(fs, 'chmod'):
      fs.chmod(path, mode)

  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    return fs.exists(path)

  def glob(self, pattern: Union[str, os.PathLike[str]]) -> list[str]:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(pattern)
    protocol = fsspec.utils.get_protocol(pattern)
    return [f'{protocol}:///{r.lstrip("/")}' for r in fs.glob(path)]

  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    return [os.path.basename(f) for f in fs.ls(path, detail=False)]

  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    return fs.isdir(path)

  def mkdir(
      self,
      path: Union[str, os.PathLike[str]], mode: int = 0o777
  ) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    fs.mkdir(path)

  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = True,
  ) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    fs.makedirs(path, exist_ok=exist_ok)

  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    fs.rm(path)

  def rename(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, old_path = fsspec.core.url_to_fs(oldpath)
    fs2, new_path = fsspec.core.url_to_fs(newpath)
    if fs.__class__ != fs2.__class__:
      raise ValueError(
          f'Rename across different filesystems is not supported: '
          f'{type(fs)} vs {type(fs2)}'
      )
    fs.rename(old_path, new_path)

  def rmdir(self, path: Union[str, os.PathLike[str]]) -> None:   # pytype: disable=signature-mismatch
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    fs.rmdir(path)

  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs, path = fsspec.core.url_to_fs(path)
    fs.rm(path, recursive=True)

  def copy(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    assert fsspec is not None, '`fsspec` is not installed.'
    fs1, old_path_in_fs = fsspec.core.url_to_fs(oldpath)
    fs2, new_path_in_fs = fsspec.core.url_to_fs(newpath)

    if fs2.isdir(new_path_in_fs):
      new_path_in_fs = os.path.join(
          new_path_in_fs, os.path.basename(old_path_in_fs)
      )

    if fs1.__class__ == fs2.__class__:
      fs1.copy(old_path_in_fs, new_path_in_fs)
    else:
      with fs1.open(old_path_in_fs, 'rb') as f1:
        with fs2.open(new_path_in_fs, 'wb') as f2:
          while True:
            # Reads the file in 16MB chunks.
            chunk = f1.read(16 * 1024 * 1024)
            if not chunk:
              break
            f2.write(chunk)


class _FsspecUriCatcher(FileSystem):
  """File system to catch URI paths and redirect to FsspecFileSystem."""

  # Catch all paths that contains '://' but not registered by
  # available_protocols.
  _URI_PATTERN = re.compile(r'^[a-zA-Z][a-zA-Z0-9+-.]*://.*')

  def __init__(self):
    super().__init__()
    self._std_fs = StdFileSystem()
    self._fsspec_fs = FsspecFileSystem()

  def get_fs(self, path: Union[str, os.PathLike[str]]) -> FileSystem:
    if self._URI_PATTERN.match(resolve_path(path)):
      return self._fsspec_fs
    return self._std_fs

  def open(
      self, path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs
  ) -> File:
    return self.get_fs(path).open(path, mode, **kwargs)

  def chmod(self, path: Union[str, os.PathLike[str]], mode: int) -> None:
    self.get_fs(path).chmod(path, mode)

  def exists(self, path: Union[str, os.PathLike[str]]) -> bool:
    return self.get_fs(path).exists(path)

  def glob(self, pattern: Union[str, os.PathLike[str]]) -> list[str]:
    return self.get_fs(pattern).glob(pattern)

  def listdir(self, path: Union[str, os.PathLike[str]]) -> list[str]:
    return self.get_fs(path).listdir(path)

  def isdir(self, path: Union[str, os.PathLike[str]]) -> bool:
    return self.get_fs(path).isdir(path)

  def mkdir(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777
  ) -> None:
    self.get_fs(path).mkdir(path, mode)

  def mkdirs(
      self,
      path: Union[str, os.PathLike[str]],
      mode: int = 0o777,
      exist_ok: bool = True,
  ) -> None:
    self.get_fs(path).mkdirs(path, mode, exist_ok)

  def rm(self, path: Union[str, os.PathLike[str]]) -> None:
    self.get_fs(path).rm(path)

  def rename(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    self.get_fs(oldpath).rename(oldpath, newpath)

  def rmdir(self, path: Union[str, os.PathLike[str]]) -> None:   # pytype: disable=signature-mismatch
    self.get_fs(path).rmdir(path)

  def rmdirs(self, path: Union[str, os.PathLike[str]]) -> None:
    self.get_fs(path).rmdirs(path)

  def copy(
      self,
      oldpath: Union[str, os.PathLike[str]],
      newpath: Union[str, os.PathLike[str]],
  ) -> None:
    self.get_fs(oldpath).copy(oldpath, newpath)


if fsspec is not None:
  fsspec_fs = FsspecFileSystem()
  for p in fsspec.available_protocols():
    add_file_system(p + '://', fsspec_fs)
  add_file_system('', _FsspecUriCatcher())


#
# APIs for file IO.
#


def open(path: Union[str, os.PathLike[str]], mode: str = 'r', **kwargs) -> File:  # pylint:disable=redefined-builtin
  """Opens a file with a path."""
  return _fs.get(path).open(path, mode, **kwargs)


def chmod(path: Union[str, os.PathLike[str]], mode: int) -> None:
  """Changes the permission of a file."""
  _fs.get(path).chmod(path, mode)


def readfile(
    path: Union[str, os.PathLike[str]],
    mode: str = 'r',
    nonexist_ok: bool = False,
    **kwargs,
) -> Union[bytes, str, None]:
  """Reads content from a file."""
  try:
    with _fs.get(path).open(path, mode=mode, **kwargs) as f:
      return f.read()
  except FileNotFoundError as e:
    if nonexist_ok:
      return None
    raise e


def writefile(
    path: Union[str, os.PathLike[str]],
    content: Union[str, bytes],
    *,
    mode: str = 'w',
    perms: Optional[int] = 0o664,   # Default to world-readable.
    **kwargs,
) -> None:
  """Writes content to a file."""
  with _fs.get(path).open(path, mode=mode, **kwargs) as f:
    f.write(content)
  if perms is not None:
    chmod(path, perms)


def rename(
    oldpath: Union[str, os.PathLike[str]],
    newpath: Union[str, os.PathLike[str]],
) -> None:
  """Renames a file."""
  _fs.get(oldpath).rename(oldpath, newpath)


def rm(path: Union[str, os.PathLike[str]]) -> None:
  """Removes a file."""
  _fs.get(path).rm(path)


def copy(
    oldpath: Union[str, os.PathLike[str]],
    newpath: Union[str, os.PathLike[str]],
) -> None:
  """Copies a file."""
  _fs.get(oldpath).copy(oldpath, newpath)


def path_exists(path: Union[str, os.PathLike[str]]) -> bool:
  """Returns True if path exists."""
  return _fs.get(path).exists(path)


def glob(pattern: Union[str, os.PathLike[str]]) -> list[str]:
  """Lists all files or sub-directories."""
  return _fs.get(pattern).glob(pattern)


def listdir(
    path: Union[str, os.PathLike[str]], fullpath: bool = False
) -> list[str]:  # pylint: disable=redefined-builtin
  """Lists all files or sub-directories under a dir."""
  entries = _fs.get(path).listdir(path)
  if fullpath:
    return [os.path.join(path, entry) for entry in entries]
  return entries


def isdir(path: Union[str, os.PathLike[str]]) -> bool:
  """Returns True if path is a directory."""
  return _fs.get(path).isdir(path)


def mkdir(path: Union[str, os.PathLike[str]], mode: int = 0o777) -> None:
  """Makes a directory."""
  _fs.get(path).mkdir(path, mode=mode)


def mkdirs(
    path: Union[str, os.PathLike[str]],
    mode: int = 0o777,
    exist_ok: bool = True,
) -> None:
  """Makes a directory chain."""
  _fs.get(path).mkdirs(path, mode=mode, exist_ok=exist_ok)


def rmdir(path: Union[str, os.PathLike[str]]) -> bool:
  """Removes a directory."""
  return _fs.get(path).rmdir(path)


def rmdirs(path: Union[str, os.PathLike[str]]) -> bool:
  """Removes a directory chain until a parent directory is not empty."""
  return _fs.get(path).rmdirs(path)  # pytype: disable=bad-return-type



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Automatic type conversions."""

import calendar
import datetime
import sys
from typing import Any, Callable, Optional, Tuple, Type, Union

from pyglove.core import utils
from pyglove.core.typing import inspect as pg_inspect


class _TypeConverterRegistry:
  """Type converter registry."""

  _JSON_VALUE_TYPES = frozenset(
      [int, float, bool, type(None), list, tuple, dict, str]
  )

  def __init__(self):
    """Constructor."""
    self._converter_list = []

  def register(
      self,
      src: Union[Type[Any], Tuple[Type[Any], ...]],
      dest: Union[Type[Any], Tuple[Type[Any], ...]],
      convert_fn: Callable[[Any], Any]) -> None:  # pyformat: disable pylint: disable=line-too-long
    """Register a converter from src type to dest type."""
    if (
        not isinstance(src, (tuple, type))
        and not pg_inspect.is_generic(src)
        or not isinstance(dest, (tuple, type))
        and not pg_inspect.is_generic(dest)
    ):
      raise TypeError('Argument \'src\' and \'dest\' must be a type or '
                      'tuple of types.')
    if isinstance(dest, tuple):
      json_value_convertible = any(d in self._JSON_VALUE_TYPES for d in dest)
    else:
      json_value_convertible = dest in self._JSON_VALUE_TYPES
    self._converter_list.append((src, dest, convert_fn, json_value_convertible))

  def get_converter(
      self, src: Type[Any], dest: Type[Any]) -> Optional[Callable[[Any], Any]]:
    """Get converter from source type to destination type."""
    if pg_inspect.is_protocol(dest):
      return None

    # TODO(daiyip): Right now we don't see the need of a large number of
    # converters, thus its affordable to iterate the list.
    # We may consider more efficient way to do lookup in future.
    # NOTE(daiyip): We do reverse lookup since usually subclass converter
    # is register after base class.
    for src_type, dest_type, converter, _ in reversed(self._converter_list):
      if pg_inspect.is_subclass(src, src_type):
        dest_types = dest_type if isinstance(dest_type, tuple) else (dest_type,)
        for dest_type in dest_types:
          if pg_inspect.is_subclass(dest_type, dest):
            return converter
    return None

  def get_json_value_converter(
      self, src: Type[Any]) -> Optional[Callable[[Any], Any]]:
    """Get converter from source type to a JSON simple type."""
    for src_type, _, converter, json_value_convertible in reversed(
        self._converter_list):
      if pg_inspect.is_subclass(src, src_type) and json_value_convertible:
        return converter
    return None


_TYPE_CONVERTER_REGISTRY = _TypeConverterRegistry()


def get_converter(
    src: Type[Any], dest: Union[Type[Any], Tuple[Type[Any], ...]]
) -> Optional[Callable[[Any], Any]]:
  """Get converter from source type to destination type."""
  dest_types = dest if isinstance(dest, tuple) else (dest,)
  for dest in dest_types:
    converter = _TYPE_CONVERTER_REGISTRY.get_converter(src, dest)
    if converter is not None:
      return converter
  return None


def get_json_value_converter(src: Type[Any]) -> Optional[Callable[[Any], Any]]:
  """Get converter from source type to a JSON simple type."""
  return _TYPE_CONVERTER_REGISTRY.get_json_value_converter(src)


def register_converter(
    src_type: Union[Type[Any], Tuple[Type[Any], ...]],
    dest_type: Union[Type[Any], Tuple[Type[Any], ...]],
    convert_fn: Callable[[Any], Any]) -> None:
  """Register converter from source type to destination type.

  Examples::

    # Add converter from int to float.
    pg.typing.register_converter(int, float, float)

    assert pg.typing.Float().apply(1) is 1.0

    # Add converter from a dict to class A.
    def from_dict(d):
      return A(**d)

    assert isinstance(pg.typing.Object(A).apply({'x': 1, 'y': 2}), A)

  Args:
      src_type: Source value type.
      dest_type: Target value type.
      convert_fn: Function that performs the conversion, in signature
        (src_type) -> dest_type.
  """
  _TYPE_CONVERTER_REGISTRY.register(src_type, dest_type, convert_fn)


def _register_builtin_converters():
  """Register built-in converters."""
  # int => float.
  register_converter(int, float, float)

  # int <=> datetime.datetime.
  if sys.version_info >= (3, 11):
    register_converter(
        int,
        datetime.datetime,
        lambda x: datetime.datetime.fromtimestamp(x, datetime.UTC)
    )
  else:
    register_converter(
        int, datetime.datetime, datetime.datetime.utcfromtimestamp
    )

  register_converter(
      datetime.datetime,
      int,
      lambda x: calendar.timegm(x.timetuple())
  )

  # string <=> KeyPath.
  register_converter(str, utils.KeyPath, utils.KeyPath.parse)
  register_converter(utils.KeyPath, str, lambda x: x.path)


_register_builtin_converters()
utils.JSONConvertible.TYPE_CONVERTER = get_json_value_converter
